;; Generated package description from pulsar.el  -*- no-byte-compile: t -*-
(define-package "pulsar" "0.2.0" "Pulse highlight on demand or after select functions" '((emacs "27.1")) :commit "4cfc18a0093b1c4478fc5dc0342f07a4f0033c8b" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "pulse" "highlight") :url "https://protesilaos.com/emacs/pulsar")
